
// Main GRID_MPI library module

int DEBUG_LEVEL  = 0 ; // Define this variable to see debug messages

#include "../Shared/multi_os.h"

#ifdef _WIN32
  #pragma warning(disable:4786)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <string>

#include "grid_mpi.h"
#include "types.h"
#include "nodeinfo.h"
#include "../Shared/net.h"
#include "../Shared/scripts.h"

#ifdef _WIN32
  #include <io.h>
  #include <winsock2.h>
  #include <direct.h>
#else
  #include <unistd.h>
  #include <sys/socket.h>
#endif


using namespace std ;

NodeInfo nodeInfo;
PacketList recvPackets;

///////////////////////////////////////////////////////////////////////////////
///////          MPI functions implemantation

///////////////////////////////////////////////////////////////////////////////
// MPI_Init
int MPG_Init(int *argc,char ***argv)
{
  int res;
  char dir[MAX_DIRECTORY_LENGTH];
  struct Packet pack;
//  char buf[255] = "";

  debug(1, "MPG_Init started\n");
  res = MPI_Init(argc,argv);
  if (res != MPI_SUCCESS) return res;

  res = MPI_Comm_size(MPI_COMM_WORLD, &nodeInfo.localSize);
  if (res != MPI_SUCCESS) return res;
  
  res = MPI_Comm_rank(MPI_COMM_WORLD,&nodeInfo.localRank);
  if (res != MPI_SUCCESS) return res;
  if (!initNetwork()) exit(1);

  pack.data = NULL;

  getcwd(dir, MAX_DIRECTORY_LENGTH);
  debug(3, "Current directory is %s\n", dir);
  
  // Read argc from the jobmanager

  //check args in env vars
  if ((getenv("_GMPI_SPEC_FILE") != NULL) && (getenv("_GMPI_SPEC_NUM") != NULL))
  {
    nodeInfo.subjobNum = atoi(getenv("_GMPI_SPEC_NUM"));
    if (!nodeInfo.init(getenv("_GMPI_SPEC_FILE")))
      terminateLibrary();
  }
  else // read from command line arguments
  {
    if (*argc < 3)
    {
      printf("ERROR. Command line parameter expected.\nSyntax: Executable <ProjectFilename> <Num>\nProjectFilename - name of a project file\nNum - a number, that points to what server in the project decription file this worker should connect. \n");
      terminateLibrary();
    }
    else
    {
      nodeInfo.subjobNum = atoi((*argv)[2]);

      if (!nodeInfo.init((*argv)[1]))
        terminateLibrary();
    }
  }

  // make connection with the gateway
  debug(3, "Client started\n");
  nodeInfo.soc = connectToServer((char *)nodeInfo.pInfo.serverList[nodeInfo.subjobNum].internalHostName.c_str(),
                                 nodeInfo.pInfo.serverList[nodeInfo.subjobNum].gatewayPort, 
                                 nodeInfo.globalRank, nodeInfo.globalRank, SERVICE_CLIENT);

  debug(3, "TotalServers = %d, globalrank = %d\n", nodeInfo.pInfo.serverList.size(), nodeInfo.globalRank);

  if (nodeInfo.soc == -1)
  {
    printf("Error. Can't connect with the server %s:%d\n",
           nodeInfo.pInfo.serverList[nodeInfo.subjobNum].internalHostName.c_str(),
           nodeInfo.pInfo.serverList[nodeInfo.subjobNum].gatewayPort);
    terminateLibrary();
  }; 
  debug(3, "Connected with the server %s:%d. Socket %d \n",nodeInfo.pInfo.serverList[nodeInfo.subjobNum].internalHostName.c_str(),
          nodeInfo.pInfo.serverList[nodeInfo.subjobNum].gatewayPort, nodeInfo.soc);

  debug(1, "MPG_Init finished\n");
  return MPI_SUCCESS;
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Comm_size(MPI_Comm comm, int *size)
{
 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;
  *size = nodeInfo.globalSize;
  return MPI_SUCCESS;
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Comm_rank(MPI_Comm comm, int *rank)
{
//  if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;
  *rank = nodeInfo.globalRank;
  return MPI_SUCCESS;
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Send(void *data,int size, MPI_Datatype type,int address,int tag,MPI_Comm comm)
{
  int res;
  int dataType, dataTypeSize;

 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;

  debug(1, "MPG_Send started\n");
  
  if (tag < 0) 
  {
    printf("ERROR. MPI_Send(). tag in MPI_Send call can't be less then 0.\n");
    terminateLibrary();
  };
  if (address < 0) 
  {
    printf("ERROR. MPI_Send(). rank in MPI_Send call can't be less then 0.\n");
    terminateLibrary();
  };

  if (isLocal(address))
    return MPI_Send(data, size, type, getLocalAddr(address), tag, MPI_COMM_WORLD); 

  if (!convertMPI_Grid(type, dataType, dataTypeSize))
    terminateLibrary();

  res = writePacket(nodeInfo.soc, genPacket(address, nodeInfo.globalRank, tag, MES_P2P, SIG_NONE, MES_MPI_SEND,
                                            dataTypeSize * size, dataType, dataTypeSize, (char *)data));
  if (res == -1)
  {
    printf("ERROR. MPG_Send(). Can't write packet.\n");
    terminateLibrary();
  }
    
  debug(1, "MPG_Send finished\n");  
  return MPI_SUCCESS; 
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Recv(void *data, int size, MPI_Datatype type, int address, int tag,
             MPI_Comm comm, MPI_Status *status)
{
 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;

  debug(1, "MPG_Recv started\n");
  if (isLocal(address))
    return MPI_Recv(data, size, type, getLocalAddr(address), tag, MPI_COMM_WORLD, status); 

  getMessage(data, size, type, address, tag, MES_MPI_SEND);
  
  if ((status != MPI_STATUS_IGNORE) && (status != MPI_STATUSES_IGNORE))
  {
    status->MPI_SOURCE = address;
    status->MPI_TAG    = tag;
    status->MPI_ERROR  = MPI_SUCCESS;
  }

  debug(1, "MPG_Recv finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Sendrecv(void *sBuf, int sCount, MPI_Datatype sType, int dest, int sTag,
                 void *rBuf, int rCount, MPI_Datatype rType, int source,
                 int rTag, MPI_Comm comm, MPI_Status *status)
{
  int res;

//  if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;
  debug(1, "MPG_Sendrecv finished\n");
  
  if (isLocal(dest) && isLocal(source))
  {
    return MPI_Sendrecv(sBuf, sCount, sType, getLocalAddr(dest), sTag, rBuf, rCount,
                        rType, getLocalAddr(source), rTag, MPI_COMM_WORLD, status);
  }

  if (isLocal(dest) && !isLocal(source))
  {
    res = MPI_Send(sBuf, sCount, sType, getLocalAddr(dest), sTag, MPI_COMM_WORLD);
    if (res != MPI_SUCCESS) return res;

    res = MPG_Recv(rBuf, rCount, rType, source, rTag, MPI_COMM_WORLD, status);
    if (res != MPI_SUCCESS) return res;
  };

  if (!isLocal(dest) && isLocal(source))
  {
    res = MPG_Send(sBuf, sCount, sType, dest, sTag, MPI_COMM_WORLD);
    if (res != MPI_SUCCESS) return res;

    res = MPI_Recv(rBuf, rCount, rType, getLocalAddr(source), rTag, MPI_COMM_WORLD, status);
    if (res != MPI_SUCCESS) return res;
  };

  if (!isLocal(dest) && !isLocal(source))
  {
    res = MPG_Send(sBuf, sCount, sType, dest, sTag, MPI_COMM_WORLD);
    if (res != MPI_SUCCESS) return res;

    res = MPG_Recv(rBuf, rCount, rType, source, rTag, MPI_COMM_WORLD, status);
    if (res != MPI_SUCCESS) return res;
  };

  debug(1, "MPG_Recv finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Bcast (void *data, int count, MPI_Datatype type, int root, MPI_Comm comm)
{
  int res;
  int dataType, dataTypeSize;
  
 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;
  debug(1, "MPG_Bcast started\n"); 
  
  debug(3,"%d, in Bcast, size = %d\n", nodeInfo.globalRank, nodeInfo.pInfo.serverList.size());
 
  if (nodeInfo.pInfo.serverList.size() == 1)
    return MPI_Bcast(data, count, type, nodeInfo.localRank, MPI_COMM_WORLD);

  if (nodeInfo.globalRank == root) // root process. send
  {
    debug(4, "%d, in Bcast root \n", nodeInfo.globalRank);
  
    if (!convertMPI_Grid(type, dataType, dataTypeSize))
      terminateLibrary();

    res = writePacket(nodeInfo.soc,genPacket(0, nodeInfo.globalRank, 0, MES_BROADCAST, SIG_NONE, MES_MPI_BCAST,
                                   dataTypeSize * count, dataType, dataTypeSize, (char *)data));
    if (res == -1) 
    {
      printf("ERROR. MPI_Bcast(). Can't write packet.\n");
      terminateLibrary();
    }

    // Send to all nodes in the local cluster and receive a copy itself
    res = MPI_Bcast(data, count, type, nodeInfo.localRank, MPI_COMM_WORLD);
    if (res != MPI_SUCCESS) 
    {
      debug(1, "MPG_Bcast finished\n");
      return res;
    }
  }
  else // all other processes. receive
  {
    debug(4, "%d, in Bcast receive \n", nodeInfo.globalRank);
  
    if (isLocal(root))
    {
      res = MPI_Bcast(data, count, type, getLocalAddr(root), MPI_COMM_WORLD);
      debug(1, "MPG_Bcast finished\n");
      return res;
    }
    else
    {
      if (nodeInfo.localRank == 0)
      {
        getMessage(data, count, type, MPI_ANY_SOURCE, MPI_ANY_TAG, MES_MPI_BCAST);
        res = MPI_Bcast(data, count, type, 0, MPI_COMM_WORLD);
        debug(1, "MPG_Bcast finished\n");
        return MPI_SUCCESS;
      }
      else
      {
        res = MPI_Bcast(data, count, type, 0, MPI_COMM_WORLD);
	debug(1, "MPG_Bcast finished\n");
        return res;
      }
    };

  };

  debug(1, "MPG_Bcast finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Barrier(MPI_Comm comm)
{
  int res;
  PacketList pList;
  int size = 0;
  
  debug(1, "MPG_Barrier started\n");
 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;

  if (nodeInfo.pInfo.serverList.size() == 1)
  {
    debug(1, "MPG_Barrier finished\n");
    return MPI_Barrier(MPI_COMM_WORLD);
  }
    
  res = MPI_Barrier(MPI_COMM_WORLD);
  if (res != MPI_SUCCESS) return res;

  if (nodeInfo.globalRank == 0)
  {
    if (!getBroadcast(pList, 0, MPI_CHAR, MES_MPI_BARRIER, MPI_COMM_WORLD))
      terminateLibrary();

    res = writePacket(nodeInfo.soc,genPacket(0, nodeInfo.globalRank, 0, MES_BROADCAST, SIG_NONE, MES_MPI_BARRIER,
                      0, TYPE_CHAR, sizeof(char), NULL));
    if (res == -1) 
    {
      printf("ERROR. MPG_Barrier. Can't write packet.\n");
      terminateLibrary();
    }
  }
  else
  {
    if (nodeInfo.localRank == 0)
    {
      res = writePacket(nodeInfo.soc,genPacket(0, nodeInfo.globalRank, 0, MES_P2P, SIG_NONE, MES_MPI_BARRIER,
                        0, TYPE_CHAR, sizeof(char), NULL));
      if (res == -1) 
      {
        printf("ERROR. MPG_Barrier. Can't write packet.\n");
        terminateLibrary();
      }

      res = getMessage(NULL, size, MPI_CHAR, 0, MPI_ANY_TAG, MES_MPI_BARRIER);
      if (res == -1) terminateLibrary();
    }
  }

  res = MPI_Barrier(MPI_COMM_WORLD);
  if (res != MPI_SUCCESS) return res;

  debug(1, "MPG_Barrier finished\n");
  return MPI_SUCCESS;
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Allgather(void *send, int sCount, MPI_Datatype sType, void *recv, int rCount, MPI_Datatype rType, MPI_Comm comm)
{
  int res;
  int dataType, dataTypeSize;
  PacketList pList;
  int i;

  debug(1, "MPG_Allgather started\n");
  
  if (nodeInfo.pInfo.serverList.size() == 1)
  {
    debug(1, "MPG_Allgather finished\n");
    return MPI_Allgather(send, sCount, sType, recv, rCount, rType, MPI_COMM_WORLD);
  }

  if (sType != rType)
  {
      printf("ERROR. MPI_Allgather. Different SendType and RecvType are not supported in the GRID. \n");
      terminateLibrary();
  }
  if (sCount != rCount)
  {
      printf("ERROR. MPI_Allgather. Different SendCount and RecvCount are not supported in the GRID. \n");
      terminateLibrary();
  }

  if (!convertMPI_Grid(sType, dataType, dataTypeSize))
    terminateLibrary();

  MPI_Gather(send, sCount, sType,recv, rCount, rType,0, MPI_COMM_WORLD);// gather in local cluster, gather in root = 0;

  if (nodeInfo.globalRank == 0)
  {
    int size = -1;
    // gather from all local roots
    if (!getBroadcast(pList, size, rType, MES_MPI_ALLGATHER, MPI_COMM_WORLD))
      terminateLibrary();
      
    debug(4,"%d, after getBroadcast, size = %d\n", nodeInfo.globalRank, nodeInfo.pInfo.serverList.size());

    char *pos = (char *)recv;
//    debug(4,"Before memcpy. pos = %p recv= %p end recv =%p, rcount = %d, dsize = %d, %d %d %d %d\n", pos, recv, 
//         (char *)pos + 8832*8, rCount, dataTypeSize, nodeInfo.localSize,pList.size(),
//	  pList[0].header.size, pList[0].header.dataTypeSize);
    
    pos += (rCount * dataTypeSize) * nodeInfo.localSize;
    for (i = 0; i< pList.size(); i++)
    {
      memcpy(pos, pList[i].data, pList[i].header.size);
      pos += pList[i].header.size;
    };

    debug(4,"%d, after getBroadcast.Mem., size = %d\n", nodeInfo.globalRank, nodeInfo.pInfo.serverList.size());

    clearPacketList(pList);
  }
  else
  {
    if (nodeInfo.localRank == 0)  // local roots
    {
      res = writePacket(nodeInfo.soc, genPacket(0, nodeInfo.globalRank, 0, MES_P2P, SIG_NONE, MES_MPI_ALLGATHER,
                                                dataTypeSize * rCount * nodeInfo.localSize, dataType, dataTypeSize, (char *)recv));
      if (res == -1) 
      {
        printf("ERROR. MPG_Allgather. Can't write packet.\n");
        terminateLibrary();
      }
    }
  };

  debug(4,"%d, befor bcast, size = %d\n", nodeInfo.globalRank, nodeInfo.pInfo.serverList.size());
  
  MPG_Bcast(recv, nodeInfo.globalSize * sCount , sType, 0, MPI_COMM_WORLD);
  debug(4, "%d, end bcast\n", nodeInfo.globalRank);

  debug(1, "MPG_Allgather finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Reduce(void *in, void *out, int size, MPI_Datatype type, MPI_Op op, int root, MPI_Comm comm)
{
  int res;
  int dataType, dataTypeSize;
  PacketList pList;
  Packet pack;
  void *tmpBuf = NULL;

  debug(1, "MPG_Reduce started\n");
//  if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;

  if (nodeInfo.pInfo.serverList.size() == 1)
  {
    debug(1, "MPG_Reduce finished\n");
    return MPI_Reduce(in, out, size, type, op, root, MPI_COMM_WORLD);
  }

  if (!convertMPI_Grid(type, dataType, dataTypeSize))
    terminateLibrary();

  tmpBuf = malloc(size * dataTypeSize);

  MPI_Reduce(in, tmpBuf, size, type, op, 0, MPI_COMM_WORLD); // reduce in local cluster, gather in root = 0;

  if (nodeInfo.globalRank == 0)
  {
    // gather from all local roots
    if (!getBroadcast(pList, size, type, MES_MPI_REDUCE, MPI_COMM_WORLD))
      terminateLibrary();

    pack.data = (char *)tmpBuf;
    pList.push_back(pack);
    
    doOper(pList, op, out);

    pList.pop_back();
    clearPacketList(pList);

    res = writePacket(nodeInfo.soc, genPacket(root, nodeInfo.globalRank, 0, MES_P2P, SIG_NONE, MES_MPI_REDUCE,
                                              dataTypeSize * size, dataType, dataTypeSize, (char *)out));
    if (res == -1) 
    {
      printf("ERROR. MPG_Reduce. Can't write packet.\n");
      terminateLibrary();
    }
  }
  else
  {
    if (nodeInfo.localRank == 0)
    {
      res = writePacket(nodeInfo.soc, genPacket(0, nodeInfo.globalRank, 0, MES_P2P, SIG_NONE, MES_MPI_REDUCE,
                                                dataTypeSize * size, dataType, dataTypeSize, (char *)tmpBuf));
      if (res == -1) 
      {
        printf("ERROR. MPG_Reduce. Can't write packet.\n");
        terminateLibrary();
      }
    }
  };

  if (nodeInfo.globalRank == root)
  {
    res = getMessage(out, size, type, 0, MPI_ANY_TAG, MES_MPI_REDUCE);
    if (res == -1) terminateLibrary();
  }

  free(tmpBuf);

  debug(1, "MPG_Reduce finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Allreduce(void *in, void *out, int size, MPI_Datatype type, MPI_Op op, MPI_Comm comm)
{
  int res;
  int dataType, dataTypeSize;
  PacketList pList;
  Packet pack;
  char *tmpBuf = NULL;

  debug(1, "MPG_AllReduce started\n");
 // if (comm != MPI_COMM_WORLD) return MPI_ERR_ARG;

  if (nodeInfo.pInfo.serverList.size() == 1)
  {
    debug(1, "MPG_Allreduce finished\n");
    return MPI_Allreduce(in, out, size, type, op,MPI_COMM_WORLD);
  }

  if (!convertMPI_Grid(type, dataType, dataTypeSize))
    terminateLibrary();

  tmpBuf = new char[size * dataTypeSize];
  MPI_Reduce(in, tmpBuf, size, type, op, 0, MPI_COMM_WORLD); // reduce in local cluster, gather in root = 0;

  if (nodeInfo.globalRank == 0)
  {
    // gather from all local roots
    
    if (!getBroadcast(pList, size, type, MES_MPI_REDUCE, MPI_COMM_WORLD))
      terminateLibrary();

    pack.data = (char *)tmpBuf;
    pack.header.dataType = dataType; 
    pack.header.dataTypeSize = dataTypeSize;
    pack.header.size = size * dataTypeSize;
    
    pList.push_back(pack);

    debug(4, "------------- size = %d %d op = %d %d \n",size, pList[0].header.dataType, op, dataTypeSize );
    fflush(stdout);   
    
    doOper(pList, op, out);
    
    pList.pop_back();
    clearPacketList(pList);
  }
  else
  {
    if (nodeInfo.localRank == 0)
    {
      res = writePacket(nodeInfo.soc, genPacket(0, nodeInfo.globalRank, 0, MES_P2P, SIG_NONE, MES_MPI_REDUCE,
                                                dataTypeSize * size, dataType, dataTypeSize, (char *)tmpBuf));
      if (res == -1) 
      {
        printf("ERROR. MPG_Allreduce. Can't write packet.\n");
        terminateLibrary();
      }
    }
  };

  MPG_Bcast(out, size, type, 0, MPI_COMM_WORLD);
  delete[] tmpBuf;
  
  debug(1, "MPG_Allreduce finished\n");
  return MPI_SUCCESS;    
};

///////////////////////////////////////////////////////////////////////////////
int MPG_Finalize()
{
  int res;

  debug(1, "MPG_Finalize started\n");

  // Send sutdown signal to gateway
  res = writePacket(nodeInfo.soc, genPacket(0, nodeInfo.globalRank, 0, 0, SIG_CLIENT_FINISHED, MES_NONE, 0, 0, 0, NULL));
  if (res == -1) printf("GMPI Warning.Can't send shutdown signal to the gateway\n");
  
  clearAll();

  debug(1, "MPG_Finalize finished\n");
  return MPI_SUCCESS;  
};


///////////////////////////////////////////////////////////////////////////////
// Service functions

///////////////////////////////////////////////////////////////////////////////
// Checks if address belongs to local space
int isLocal(int address)
{
  int start = 0;

  for (int i =0; i< nodeInfo.subjobNum; i++)
  {
    start+= nodeInfo.pInfo.serverList[i].nProcesses;
  }

  if ((address >= start) && (address < start + nodeInfo.localSize)) return 1;
  else return 0;
};

///////////////////////////////////////////////////////////////////////////////
int getLocalAddr(int globalAddr)
{
  int start = 0;

  if (!isLocal(globalAddr)) return -1;

  for (int i =0; i< nodeInfo.subjobNum; i++)
  {
    start+= nodeInfo.pInfo.serverList[i].nProcesses;
  }

  return (globalAddr - start) ;
};

///////////////////////////////////////////////////////////////////////////////
// return temporary list of all local roots ranks in all subjobs
IntList getAllRoots(MPI_Comm comm)
{
  static IntList rootList;
  int base = 0;

  rootList.clear();

  for (int i = 0; i< nodeInfo.pInfo.serverList.size(); i++)
  {
    rootList.push_back(base);
    base += nodeInfo.pInfo.serverList[i].nProcesses;
  }

  return rootList;
};

///////////////////////////////////////////////////////////////////////////////
// clear all data and close network
void clearAll()
{
  if (nodeInfo.soc != -1)
  {
    #ifdef _WIN32
      shutdown(nodeInfo.soc,SD_SEND);
      closesocket(nodeInfo.soc);
    #else
      shutdown(nodeInfo.soc,SHUT_WR);
      close(nodeInfo.soc);
    #endif
  }

  closeNetwork();
  recvPackets.clear();

  MPI_Finalize();
  
};

///////////////////////////////////////////////////////////////////////////////
// terminate GRID_MPI library
void terminateLibrary()
{
  printf("GRID_MPI library terminated.\n");
  clearAll();

  exit(1);
};

///////////////////////////////////////////////////////////////////////////////
int processSignal(int sig)
{
  switch (sig)
  {
    case SIG_TERMINATE     :printf("Shutdown signal recieved from jobmanager\n");
                            terminateLibrary();
    case SIG_NONE          :return 1;

    default                :printf("ERROR. Unknown signal %d recived\n", sig);
                            break;
  };

  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// receive specified message through gateway (with memory allocation )
void *getMessageAlloc(int &size, MPI_Datatype &type,int address, int tag, int mesSender)
{
  int res;
  struct Packet pack;      
  PacketList::iterator i;
  int recvDataSize;
  MPI_Datatype recvMPI_type;
  void *data = NULL;

  debug(2, "getMessageAlloc started\n");

// receive through gateway
  while (true)
  {
    // check if the packet was already received
    for (i = recvPackets.begin(); i != recvPackets.end(); i++)
    {
      if ((address == MPI_ANY_SOURCE) || (i->header.sendAddr == address))
        if ((tag == MPI_ANY_TAG) || (i->header.tag == tag))
          if (mesSender == i->header.mesSender)
          {
            if ( !convertGrid_MPI(i->header.dataType, recvMPI_type, recvDataSize) )
              terminateLibrary(); 

            if (recvMPI_type != type)
            {
              printf("WARNING! Data types don't match in getMessageAlloc call\n");
            };
            if (recvDataSize != i->header.dataTypeSize)
            {
              printf("WARNING! Sizes of data types don't match in getMessageAlloc call (grid type = %d, size on this PC = %d bytes, size received = %d bytes)\n",
	                        i->header.dataType, recvDataSize, i->header.dataTypeSize);
            }

            data = new char[i->header.size * recvDataSize];
            size = i->header.size;

            memcpy(data, i->data, i->header.size);
            free(i->data);
            recvPackets.erase(i);
	    
            debug(2, "getMessageAlloc finished\n");
            return data;
          }
          else
          {
            printf("ERROR.MPI. getMessage. Message was sent by %d and trying to receive with %d\n",
                   i->header.mesSender, mesSender );
            terminateLibrary(); 
          };
    }

    // receive new packets until we find what we need
    pack.data = NULL; // to avoid free() in next readPacketAlloc
    res = readPacketAlloc(nodeInfo.soc,&pack);
    if (res == -1)
    {
      printf("ERROR. getMessageAlloc. Error while reading a new packet (readPacketAlloc)\n");
      terminateLibrary(); 
    }

    processSignal(pack.header.signal);
    if (pack.header.signal == SIG_NONE)
      recvPackets.push_back(pack);
  }

};

///////////////////////////////////////////////////////////////////////////////
// receive specified message through gateway (without memory allocation )
int getMessage(void *data, int size, MPI_Datatype type,int address, int tag, int mesSender)
{
  void *resData;
  int getSize, gridType, memSize;

  debug(2, "getMessage started\n");
  convertMPI_Grid(type, gridType, memSize);
  resData = getMessageAlloc(getSize, type, address, tag, mesSender);
  if (getSize != size * memSize)
  {
    printf("ERROR. MPI_Recv. buffer to small to get so much data\n");
    return 0;
  }

  memcpy(data, resData, size * memSize);
  free(resData);
  
  debug(2, "getMessage finished\n");
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// size < 0 means to allocate memory for 'data' (we don't know the real size of a packet before it comes)
int getBroadcast(PacketList &pList, int size, MPI_Datatype type, int mesSender, MPI_Comm comm)
{
  static IntList rootList = getAllRoots(MPI_COMM_WORLD); // WARNING. This works only if comm is always the same!!

  int i;
  int res;
  struct Packet pack;      
  int dataType, dataTypeSize;
  int getSize;

  debug(2, "getBroadcast started\n");
  if (!convertMPI_Grid(type, dataType, dataTypeSize))
    terminateLibrary();

  debug(4, "rootList.size = %d\n", rootList.size());
  for (i=0; i < rootList.size(); i++)
  {
    if (isLocal(rootList[i])) continue; // skip messages from this subjob

    if (size < 0)
    {
      pack.data = (char *)getMessageAlloc(getSize, type,rootList[i], MPI_ANY_TAG, mesSender);
    }
    else
    {
      pack.data = new char[size * dataTypeSize];
      debug(4, "getBroadcast. memory allocated, size = %d\n", size * dataTypeSize);
      
      res = getMessage(pack.data, size, type,rootList[i], MPI_ANY_TAG, mesSender);
      getSize = size * dataTypeSize;
      if (!res) return 0;
    }

    pack.header.dataType = dataType; 
    pack.header.dataTypeSize = dataTypeSize;
    pack.header.mesSender = mesSender;
    pack.header.sendAddr = rootList[i];
    pack.header.size = getSize;

    pList.push_back(pack);
  };

  debug(2, "getBroadcast finished\n");
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
void clearPacketList(PacketList &pList)
{
  for (int i =0; i < pList.size(); i++)
  {
    //free(pList[i].data);
    delete[] pList[i].data;
  }

  pList.clear();
};

///////////////////////////////////////////////////////////////////////////////
void doOper(PacketList &pList, MPI_Op operation, void *out)
{
  int i, j;
  double op1f;
  long op1i;
  
  debug(2, "doOper started\n");

  for (i =0; i < pList.size(); i++)
  {
    if ((pList[0].header.size != pList[i].header.size) || (pList[0].header.dataType != pList[i].header.dataType)
        || (pList[0].header.dataTypeSize != pList[i].header.dataTypeSize))
    {
      printf("ERROR. doOper. AllReduce - diffirent data received from processes. \n");
      terminateLibrary(); 
    }
  };
  
  for (j = 0; j < (pList[0].header.size / pList[0].header.dataTypeSize); j++)
  {
    //--------------- MPI_SUM
    if (operation == MPI_SUM) 
    {
      op1i = 0;
      op1f = 0.0;
      if (j < 8) debug(4, "total %d, step %d\n", pList[0].header.size,j);

      for (i =0; i < pList.size(); i++)
      {
        switch (pList[0].header.dataType)
        {
          case TYPE_SHORT :op1i += ((short *)pList[i].data)[j];
                           break;
          case TYPE_INT   :op1i += ((int *)pList[i].data)[j];
                           break;
          case TYPE_LONG  :op1i += ((long *)pList[i].data)[j];
                           break;
          case TYPE_USHORT:op1i += ((unsigned short *)pList[i].data)[j];
                           break;
          case TYPE_ULONG :op1i += ((unsigned long *)pList[i].data)[j];
                           break;
          case TYPE_FLOAT :op1f += ((float *)pList[i].data)[j];
                           break;
          case TYPE_DOUBLE:if (j < 8) debug(4,"a"); if (j < 8) debug(4," point = %p, ",(double *)pList[i].data);
	                   if (j < 8) debug(4," val = %f ",((double *)pList[i].data)[j]);
			   
  		           op1f +=  ((double *)pList[i].data)[j];
                           break;

          default:printf("ERROR. doOper().Unsupported data type for this operation. Datatype code = %d.\n", pList[0].header.dataType);
                  terminateLibrary();
        };
      }
    }

    //--------------- MPI_PROD
    if (operation == MPI_PROD) 
    {
      op1i = 1;
      op1f = 1.0;

      for (i =0; i < pList.size(); i++)
      {
        switch (pList[0].header.dataType)
        {
          case TYPE_SHORT :op1i *= ((short *)pList[i].data)[j];
                           break;
          case TYPE_INT   :op1i *= ((int *)pList[i].data)[j];
                           break;
          case TYPE_LONG  :op1i *= ((long *)pList[i].data)[j];
                           break;
          case TYPE_USHORT:op1i *= ((unsigned short *)pList[i].data)[j];
                           break;
          case TYPE_ULONG :op1i *= ((unsigned long *)pList[i].data)[j];
                           break;
          case TYPE_FLOAT :op1f *= ((float *)pList[i].data)[j];
                           break;
          case TYPE_DOUBLE:op1f *= ((double *)pList[i].data)[j];
                           break;

          default:printf("ERROR. doOper().Unsupported data type for this operation. Datatype code = %d.\n", pList[0].header.dataType);
                  terminateLibrary();
        };
      }
    }

    //--------------- MPI_MAX
    if (operation == MPI_MAX) 
    {
      op1i = -21474836; 
      op1f = -1.7e308 ; 

      for (i =0; i < pList.size(); i++)
      {
        switch (pList[0].header.dataType)
        {
          case TYPE_SHORT :if (op1i < ((short *)pList[i].data)[j]) op1i = ((short *)pList[i].data)[j] ;
                           break;
          case TYPE_INT   :if (op1i < ((int *)pList[i].data)[j]) op1i = ((int *)pList[i].data)[j] ;
                           break;
          case TYPE_LONG  :if (op1i < ((long *)pList[i].data)[j]) op1i = ((long *)pList[i].data)[j] ;
                           break;
          case TYPE_USHORT:if (op1i < ((unsigned short *)pList[i].data)[j]) op1i = ((unsigned short *)pList[i].data)[j] ;
                           break;
          case TYPE_ULONG :if (op1i < ((unsigned long *)pList[i].data)[j]) op1i = ((unsigned long *)pList[i].data)[j] ;
                           break;
          case TYPE_FLOAT :if (op1f < ((float *)pList[i].data)[j]) op1f = ((float *)pList[i].data)[j] ;
                           break;
          case TYPE_DOUBLE:if (op1f < ((double *)pList[i].data)[j]) op1f = ((double *)pList[i].data)[j] ;
                           break;

          default:printf("ERROR. doOper().Unsupported data type for this operation. Datatype code = %d.\n", pList[0].header.dataType);
                  terminateLibrary();
        };
      }
    }

    //--------------- MPI_MIN
    if (operation == MPI_MIN) 
    {
      op1i = 2147483647; 
      op1f = 1.7e308 ; 

      for (i =0; i < pList.size(); i++)
      {
        switch (pList[0].header.dataType)
        {
          case TYPE_SHORT :if (op1i > ((short *)pList[i].data)[j]) op1i = ((short *)pList[i].data)[j] ;
                           break;
          case TYPE_INT   :if (op1i > ((int *)pList[i].data)[j]) op1i = ((int *)pList[i].data)[j] ;
                           break;
          case TYPE_LONG  :if (op1i > ((long *)pList[i].data)[j]) op1i = ((long *)pList[i].data)[j] ;
                           break;
          case TYPE_USHORT:if (op1i > ((unsigned short *)pList[i].data)[j]) op1i = ((unsigned short *)pList[i].data)[j] ;
                           break;
          case TYPE_ULONG :if (op1i > ((unsigned long *)pList[i].data)[j]) op1i = ((unsigned long *)pList[i].data)[j] ;
                           break;
          case TYPE_FLOAT :if (op1f > ((float *)pList[i].data)[j]) op1f = ((float *)pList[i].data)[j] ;
                           break;
          case TYPE_DOUBLE:if (op1f > ((double *)pList[i].data)[j]) op1f = ((double *)pList[i].data)[j] ;
                           break;

          default:printf("ERROR. doOper().Unsupported data type for this operation. Datatype code = %d.\n", pList[0].header.dataType);
                  terminateLibrary();
        };
      }
    }

    //---------------
    switch (pList[0].header.dataType)
    {
      case TYPE_SHORT :((short *)out)[j] = op1i;
                       break;
      case TYPE_INT   :((int *)out)[j] = op1i;
                       break;
      case TYPE_LONG  :((long *)out)[j] = op1i;
                       break;
      case TYPE_USHORT:((unsigned short *)out)[j] = op1i;
                       break;
      case TYPE_ULONG :((unsigned long *)out)[j] = op1i;
                       break;
      case TYPE_FLOAT :((float *)out)[j] = op1f;
                       break;
      case TYPE_DOUBLE:((double *)out)[j] =op1f;
                       break;
    }
  }

  debug(2, "doOper finished\n");
};
