
// Low level network functions

#include <errno.h>       
#include <string.h>
#include <stdio.h>

#include "../Shared/multi_os.h"
#include "net.h"

#ifdef _WIN32
  #include <winsock2.h>
  #include <io.h>
#else
  #include <unistd.h>
  #include <signal.h>
  #include <netdb.h>
  #include <sys/socket.h>
  #include <stdlib.h>
  #include <netinet/in.h>
#endif

///////////////////////////////////////////////////////////////////////////////
// Call to initialize the network library. Must be called before
// using all the other network functions
int initNetwork()
{
#ifdef _WIN32
  WORD wVersionRequested;
  WSADATA wsaData;
  int err;
 
  wVersionRequested = MAKEWORD( 2, 2 );
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) 
  {
    printf("ERROR. initNetwork(). Can't initialize sockets library\n");
    return 0;
  }
#else
  signal(SIGPIPE, SIG_IGN);
#endif
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Call to finalize network library. Must be called at the end
void closeNetwork()
{
#ifdef _WIN32
  WSACleanup();
#endif
};

///////////////////////////////////////////////////////////////////////////////
// Connect to host and port. Returns connected socket
// or -1 in case of error
int callSocket(char *hostname, unsigned short portnum)
{ 
  struct sockaddr_in sa;
  struct hostent     *hp;
  int s,res;

  if ((hp= gethostbyname(hostname)) == NULL)
  {                                                 // do we know the host's address? 
    printf("ERROR.gethostbyname(). Can't retrieve host information corresponding to a host name from a host database. Host = %s\n", hostname);
    return -1;                                      // no 
  }

  memset(&sa,0,sizeof(sa));
  memcpy((char *)&sa.sin_addr,hp->h_addr,hp->h_length);   //set address 
  sa.sin_family= hp->h_addrtype;
  sa.sin_port= htons((u_short)portnum);
  
  if ((s= socket(hp->h_addrtype,SOCK_STREAM,0)) < 0)      // get socket 
    return -1;

  res = connect(s,(struct sockaddr *)&sa,sizeof sa);
  
  if ( res <0)
  {
    char buf[20];
    sprintf(buf,"%d",portnum);
    std::string outErr = "callSocket(). Trying to connect to ";
    outErr+=hostname; outErr+=":"; outErr+=buf;
    
    showSocketError((char *)outErr.c_str());
    close(s);
    return -1;
  }
  
  return s;
}

///////////////////////////////////////////////////////////////////////////////
// Establish connection to numGRID server
// min and max ranks must be passed as arguments
// returns connected socket or -1 in case of error
int connectToServer(char *server, int port, int min, int max, int service)
{
  int soc,i;
  struct Packet pack;
  struct ConnectInfo info;

  debug(4, "connectToServer started\n");  
  for (i = 0; i < CONNECT_TRY_TIMES; i++)
  {
    soc = callSocket(server, port);
    if (soc != -1) break;
    delay(TIME_BETWEEN_CONNECT_TRY);
  }
  
  if (soc == -1) return -1;
  
  // send contact info
  info.id = PROGRAM_ID;
  info.service = service;
  info.ngVersion = NG_VERSION;
  info.minRank = min;
  info.maxRank = max;
  pack.header.size = sizeof(struct ConnectInfo);
  pack.data = (char *) &info; 
  if ( writePacket(soc,&pack) == -1) return -1;
  
  debug(4, "connectToServer finished\n");  
  return soc;
};

///////////////////////////////////////////////////////////////////////////////
// Make and bind server socket on port. Returns connected socket
// or -1 in case of error
int makeServerSocket(unsigned short portnum)
{ 
  char   myname[MAX_HOSTNAME_LENGTH+1];
  int    s;
  struct sockaddr_in sa;
  struct hostent *hp;

  memset(&sa, 0, sizeof(struct sockaddr_in)); // clear our address 

  gethostname(myname, MAX_HOSTNAME_LENGTH);   // who are we? 
  hp= gethostbyname(myname);                  // get our address info 

  if (hp == NULL)                             // we don't exist !? 
    return -1;

  sa.sin_family= hp->h_addrtype;              // this is our host address 
  sa.sin_port= htons(portnum);                // this is our port number 
  sa.sin_addr.s_addr = INADDR_ANY;

  if ((s= socket(AF_INET, SOCK_STREAM, 0)) < 0) // create socket 
  {
    showSocketError("makeServerSocket()");
    return -1;
  }
  
  int reuse = 1;
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) != 0)
  {
    showSocketError("makeServerSocket()");
    close(s);
    return -1;
  }
      
  if (bind(s,(struct sockaddr *)&sa,sizeof(struct sockaddr_in)) < 0) 
  {
#ifdef _WIN32  
    showSocketError("makeServerSocket()");
#else
    printf("Error in makeServerSocket.bind().");
    switch (errno)
    {
      case EBADF   : printf("Socket is not a valid descriptor\n"); break;
      case EINVAL  : printf("The socket is already bound to an address.\n"); break;
      case EACCES  : printf("The address is protected, and the user is not the super-user\n"); break;
      case ENOTSOCK: printf("Argument is a descriptor for a file, not a socket\n"); break;
      default      : printf("Unknown error code %d (see socket's manual)\n", errno); break;
    };
#endif
    close(s);
    return -1;                               // bind address to socket 
  }
  listen(s, 3);                               // max # of queued connects 

  return s;
}

///////////////////////////////////////////////////////////////////////////////
// Wait for new connection. Returns new socket or -1 in case of error
int acceptConnection(int soc)
{
  int newSoc;                  

  if ((newSoc = accept(soc,NULL,NULL)) < 0) 
  {
    #ifdef _WIN32
      if (WSAGetLastError() == WSAEINTR) return -1; // skip error info .this may be in case of normal close(sock)
    #endif

    showSocketError("acceptConnection()");
    return -1;
  }

  return newSoc;
}

///////////////////////////////////////////////////////////////////////////////
// Reads size bytes from connected socket into buf (must be allocated before!)
// returns number bytes read or -1 in case of error
int readData(int soc, char *data, int size)
{
  int count = 0, br =0;
  char *buf = data;

  while (count < size)
  {
    if ((br = recv(soc, buf, size - count,0)) > 0)    
    {
      count +=br;
      buf +=br;
      if (count >=size) break;
    }
    else 
    {
      if (br < 0) 
      {
        if (errno != EINTR) 
        {
          char inf[100];
          sprintf(inf,"readData (socket = %d)", soc);
          showSocketError(inf);
          return -1;
        }
      }
      else
      {
        printf("Socket error in 'readData' (socket = %d), Connection was closed by remote client. \n", soc); 
	      return -1;
      }	
    }
  }
  
  return count;       
};

///////////////////////////////////////////////////////////////////////////////
// Writes size bytes to connected socket from buf (must be allocated before!)
// returns number bytes written or -1 in case of error
int writeData(int soc, char *buf, int size)
{
  int ret;

  debug(6, "writeData started\n");
  if (size ==0) return 0;
  
  ret = send(soc,buf,size,0);
  if (ret < 0) 
    showSocketError("writeData()");

  debug(6, "writeData finished\n");
  return ret;
};

///////////////////////////////////////////////////////////////////////////////
// Reads net packet into struct p from the socket . memory MUST be allocated
// before. Returns 1 on success or -1 in case of error
int readPacket(int soc, struct Packet *p)
{
  if (readData(soc,(char *)&p->header, sizeof(struct PacketHeader)) < 0)
    return -1;
 
  // memory check
  if (p->data == NULL)
  {
    printf("Error. NULL pointer passed as an argument. (Read_Packet)\n");
    return -1;
  }
 
  if (readData(soc,p->data ,p->header.size) < 0) return -1;
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// Reads packet into struct p from the socket. memory is allocated dynamically.
// later you should free this memory manually using free.
// returns 1 on success or -1 in case of error
int readPacketAlloc(int soc, struct Packet *p)
{
  debug(5, "readPacketAlloc started\n");
  
  if (readData(soc,(char *)&p->header, sizeof(struct PacketHeader))  < 0)
    return -1;
 
  // memory alloc
  if (p->data != NULL) free(p->data);
 
  p->data = (char *)malloc(p->header.size);
  if (p->data == NULL)
  {
    printf("Error. Can't allocate memory for the buffer. (Read_PacketAlloc)\n");
    return -1;
  }
 
  if (readData(soc,p->data ,p->header.size) < 0) return -1;

  debug(5, "readPacketAlloc finished\n");
  return 1;
};
///////////////////////////////////////////////////////////////////////////////
// Writes net packet from struct p into the socket.
// returns 1 on success or -1 in case of error
int writePacket(int soc, struct Packet *p)
{
  int res;

  debug(5, "writePacket started\n");
  res = writeData(soc, (char *)&(p->header), sizeof(struct PacketHeader));
  if (res == -1) return -1;
  
  res = writeData(soc, (char *)p->data, p->header.size);
  if (res == -1) return -1; 
  
  debug(5, "writePacket finished\n");
  return 1;
};

///////////////////////////////////////////////////////////////////////////////
// generate TEMPORARY GMPI packet for 1 operation
Packet *genPacket(int recv, int send, int tag, char mesType, char signal, char mesSender, 
                  int size, int dataType, int dataTypeSize, char *data)
{
  static Packet pack;

  pack.header.dataType     = dataType;
  pack.header.dataTypeSize = dataTypeSize;
  pack.header.mesType      = mesType;
  pack.header.recvAddr     = recv;
  pack.header.sendAddr     = send;
  pack.header.signal       = signal;
  pack.header.size         = size;
  pack.header.tag          = tag;
  pack.header.mesSender    = mesSender;
  pack.data                = data;

  return &pack;
};

///////////////////////////////////////////////////////////////////////////////
// Carefully closes socket 'soc'
void closeSocket(int soc)
{
#ifdef _WIN32
  shutdown(soc, SD_SEND);
  closesocket(soc);
#else
  shutdown(soc, SHUT_WR);
  close(soc);
#endif
};

///////////////////////////////////////////////////////////////////////////////
// Shows last socket error. mes can be used to show function name where
// socket info occured
void showSocketError(char *mes)
{
  printf("Socket error in %s : ",mes);
#ifdef _WIN32
  int code = WSAGetLastError();
  switch (code)
  {
    case 0                 :printf("Socket was closed.\n"); break;
    case WSANOTINITIALISED :printf("A successful WSAStartup must occur before using this function.\n"); break;
    case WSAENETDOWN       :printf("The network subsystem has failed.\n"); break;
    case WSAEFAULT         :printf("The buf parameter is not completely contained in a valid part of the user address space.\n"); break;
    case WSAENOTCONN       :printf("The socket is not connected.\n"); break;
    case WSAEINTR          :printf("The (blocking) call was canceled through WSACancelBlockingCall.\n"); break;
    case WSAEINPROGRESS    :printf("A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function.\n"); break;
    case WSAENETRESET      :printf("The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress.\n"); break;
    case WSAENOTSOCK       :printf("The descriptor is not a socket.\n"); break;
    case WSAEOPNOTSUPP     :printf("MSG_OOB was specified, but the socket is not stream-style such as type SOCK_STREAM, out-of-band data is not supported in the communication domain associated with this socket, or the socket is unidirectional and supports only send operations.\n"); break;
    case WSAESHUTDOWN      :printf("The socket has been shut down; it is not possible to recv on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH.\n"); break;
    case WSAEWOULDBLOCK    :printf("The socket is marked as nonblocking and the receive operation would block.\n"); break;
    case WSAEMSGSIZE       :printf("The message was too large to fit into the specified buffer and was truncated.\n"); break;
    case WSAEINVAL         :printf("The socket has not been bound with bind, or an unknown flag was specified, or MSG_OOB was specified for a socket with SO_OOBINLINE enabled or (for byte stream sockets only) len was zero or negative.\n"); break;
    case WSAECONNABORTED   :printf("The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable.\n"); break;
    case WSAETIMEDOUT      :printf("The connection has been dropped because of a network failure or because the peer system failed to respond.\n"); break;
    case WSAECONNRESET     :printf("The virtual circuit was reset by the remote side executing a 'hard' or 'abortive' close. The application should close the socket as it is no longer usable. On a UDP datagram socket this error would indicate that a previous send operation resulted in an ICMP 'Port Unreachable' message.\n"); break;
    case WSAEMFILE         :printf("The queue is nonempty upon entry to accept and there are no descriptors available.\n"); break;
    case WSAENOBUFS        :printf("No buffer space is available.\n"); break;
    case WSAEADDRINUSE     :printf("The socket's local address is already in use and the socket was not marked to allow address reuse with SO_REUSEADDR. This error usually occurs when executing bind, but could be delayed until this function if the bind was to a partially wild-card address (involving ADDR_ANY) and if a specific address needs to be committed at the time of this function.\n"); break;
    case WSAEALREADY       :printf("A nonblocking connect call is in progress on the specified socket. Note In order to preserve backward compatibility, this error is reported as WSAEINVAL to Windows Sockets 1.1 applications that link to either WINSOCK.DLL or WSOCK32.DLL.\n"); break;
    case WSAEADDRNOTAVAIL  :printf("The remote address is not a valid address (such as ADDR_ANY).\n"); break;
    case WSAEAFNOSUPPORT   :printf("Addresses in the specified family cannot be used with this socket.\n"); break;
    case WSAECONNREFUSED   :printf("The attempt to connect was forcefully rejected.\n"); break;
    case WSAEISCONN        :printf("The socket is already connected (connection-oriented sockets only).\n"); break;
    case WSAENETUNREACH    :printf("The network cannot be reached from this host at this time.\n"); break;
    case WSAEACCES         :printf("Attempt to connect datagram socket to broadcast address failed because setsockopt option SO_BROADCAST is not enabled.\n"); break;

    default                :printf("Unknown error code %d (see socket's manual)\n", code); break;
  }
#else
  switch (errno)
  {
    case EPROTONOSUPPORT:printf("The  protocol  type  or  the specified protocol is not supported within this domain.\n"); break;
    case EAFNOSUPPORT:printf("The implementation does not support the specified  address  family.\n"); break;
    case ENFILE :printf("Not enough kernel memory to allocate a new socket structure.\n"); break;
    case EMFILE :printf("Process file table overflow.\n"); break;
    case EACCES :printf("Permission  to create a socket of the specified type and/or protocol is denied.\n"); break;
    case ENOBUFS:printf("Insufficient memory is available.  The socket cannot be  created until sufficient resources are freed.\n"); break;
    case ENOMEM :printf("Insufficient memory is available.  The socket cannot be  created until sufficient resources are freed.\n"); break;
    case EINVAL :printf("Unknown protocol, or protocol family not available.\n"); break;
    case ECONNREFUSED :printf("Connection refused.\n"); break;
    
    case EBADF  :printf("\n"); break;

    default     :printf("Unknown error code %d (see socket's manual)\n", errno); break;
  };
#endif
};
