// ProjectInfo.cpp: implementation of the ProjectInfo class.
//
///////////////////////////////////////////////////////////////////////////////

#pragma warning(disable:4786)

#include "projectinfo.h"
#include "projectsyntax.h"
#include "common.h"

///////////////////////////////////////////////////////////////////////////////
void ServerInfo::clear()
{
  serverPort = DEFAULT_SERVER_PORT;
  gatewayPort = DEFAULT_GATEWAY_PORT;

  blockName = "";
  serverName = "";
  internalHostName = "";
  nProcesses = -1;

  username = "";
  password = "";

  makefile = "";
  sourceDir = "";
  inputDir = "";
  outputDir = "";

  workDir = "";

  needRunTask = true;
  executable = "";
  execArgs.clear();
  execEnv.clear();

  sourceFiles.clear();
  inputFiles.clear();
  outputFiles.clear();  
};

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
ProjectInfo::ProjectInfo()
{
  serverList.clear();
  projectFile = "";
}

///////////////////////////////////////////////////////////////////////////////
ProjectInfo::~ProjectInfo()
{
}

///////////////////////////////////////////////////////////////////////////////
// Insert new variable's data into ServerInfo
int ProjectInfo::parseVar(NGVarInfo &var, ServerInfo &sInfo)
{
  int i;
  int code, parCode;

  code = getVarCode(var.name);
  if (var.params.size() == 0)
  {
    printf("Internal error.ProjectInfo::parseVar. Size = 0\n");
    return 0;
  };
  
  if (code == PAR_ERROR) 
  {
    errCode = ERR_UNKNOWN_IDENT;
    return 0;
  };

  switch(code)
  {
    case PAR_USERNAME          :sInfo.username = var.params[0]; break;
    case PAR_PASSWORD          :sInfo.password = var.params[0]; break;

    case PAR_SOURCE_DIR        :sInfo.sourceDir = var.params[0]; break;
    case PAR_INPUT_DIR         :sInfo.inputDir = var.params[0]; break;
    case PAR_OUTPUT_DIR        :sInfo.outputDir = var.params[0]; break;
    case PAR_WORK_DIR          :sInfo.workDir = var.params[0]; break;

    case PAR_EXECUTABLE        :sInfo.executable = var.params[0]; break;
    case PAR_NEED_RUN_TASK     :parCode = atoi(var.params[0].c_str());
                                switch (parCode)
                                {
                                  case 1   :sInfo.needRunTask = true; break;
                                  case 0   :sInfo.needRunTask = false; break;
                                  default  :errCode = ERR_WRONG_PARAM; return 0;
                                };
                                break;

    case PAR_SERVER_NAME       :sInfo.serverName = var.params[0]; break;
    case PAR_SERVER_PORT       :sInfo.serverPort = atoi(var.params[0].c_str());
                                if (sInfo.serverPort == 0)
                                {
                                  errCode = ERR_WRONG_VALUE;
                                  return 0;
                                };
                                break;
    case PAR_INTERNAL_HOST_NAME:sInfo.internalHostName = var.params[0]; break;
    case PAR_GATEWAY_PORT      :sInfo.gatewayPort = atoi(var.params[0].c_str());
                                if (sInfo.gatewayPort == 0)
                                {
                                  errCode = ERR_WRONG_VALUE;
                                  return 0;
                                };
                                break;
    case PAR_NP                :sInfo.nProcesses = atoi(var.params[0].c_str());
                                if (sInfo.nProcesses == 0)
                                {
                                  errCode = ERR_WRONG_VALUE;
                                  return 0;
                                };
                                break;
    case PAR_MAKEFILE          :sInfo.makefile = var.params[0]; break;
    case PAR_EXEC_ARGS         :for (i =0; i< var.params.size(); i++) 
                                  sInfo.execArgs.insert(sInfo.execArgs.end(), var.params[i]);
                                break;
    case PAR_EXEC_ENV          :for (i =0; i< var.params.size(); i++) 
                                  sInfo.execEnv.insert(sInfo.execEnv.end(), var.params[i]);
                                break;
    case PAR_INPUT_FILES       :for (i =0; i< var.params.size(); i++) 
                                  sInfo.inputFiles.insert(sInfo.inputFiles.end(), var.params[i]);
                                break;
    case PAR_OUTPUT_FILES      :for (i =0; i< var.params.size(); i++) 
                                  sInfo.outputFiles.insert(sInfo.outputFiles.end(), var.params[i]);
                                break;
    case PAR_SOURCE_FILES      :for (i =0; i< var.params.size(); i++) 
                                  sInfo.sourceFiles.insert(sInfo.sourceFiles.end(), var.params[i]);
                                break;

    default: errCode = ERR_UNKNOWN_VAR_NAME; return 0;
  };

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
// Load info from the file
int ProjectInfo::LoadInfo(char *filename)
{
  NGScript s;
  string blockName = "";
  ServerInfo sInfo;

  debug(3, "ProjectInfo::LoadInfo started\n");
  if (!s.open(filename)) 
  {
    printf("ERROR. Can't open file with Project configuration (%s)\n", filename);
    return 0;
  }

  sInfo.clear();
  
  while (true)
  {
    if (s.readVar() == NULL) // error
    {
      if (s.errCode == ERR_END_OF_FILE) break;

      printf("ERROR in %s, line %d, %s\n",filename, s.getCurLine(), s.getErrorText());
      return 0;
    };

    if (s.blockChanged)
    {
	  serverList.insert(serverList.end(), sInfo);
      sInfo.clear();
      sInfo.blockName = s.var.blockName;
    };

    if (!parseVar(s.var, sInfo))
    {
      printf("ERROR in %s, line %d, %s\n",filename, s.getCurLine() -1 , getErrorText());
      return 0;
    };
  }

  serverList.insert(serverList.end(), sInfo);
  rebuildBlocks();
  if (!checkParams()) 
  {
    printf("ERROR. %s", getErrorText());
    return 0;
  }

  debug(3, "ProjectInfo::LoadInfo finished\n");
  projectFile = filename;
  return 1;
}

///////////////////////////////////////////////////////////////////////////////
int ProjectInfo::rebuildBlocks()
{
  int i, j;

  // fill 0 block - common
  if (serverList[0].gatewayPort == -1) serverList[0].gatewayPort = DEFAULT_GATEWAY_PORT;
  if (serverList[0].serverPort == -1) serverList[0].serverPort = DEFAULT_SERVER_PORT;

  for (i = 1; i< serverList.size(); i++)
  {
    if (serverList[i].internalHostName.compare("") == 0) serverList[i].internalHostName = serverList[i].serverName;

    if (serverList[i].serverPort == -1) serverList[i].serverPort = serverList[0].serverPort;
    if (serverList[i].gatewayPort == -1) serverList[i].gatewayPort = serverList[0].gatewayPort;

    if (serverList[i].sourceDir.compare("") == 0) serverList[i].sourceDir = serverList[0].sourceDir;
    if (serverList[i].inputDir.compare("") == 0) serverList[i].inputDir = serverList[0].inputDir;
    if (serverList[i].outputDir.compare("") == 0) serverList[i].outputDir = serverList[0].outputDir;

    if (serverList[i].workDir.compare("") == 0) serverList[i].workDir = serverList[0].workDir;
    if (serverList[i].executable.compare("") == 0) serverList[i].executable = serverList[0].executable;

    if (serverList[i].makefile.compare("") == 0) serverList[i].makefile = serverList[0].makefile;

    if (serverList[i].execArgs.size() == 0) 
    {
      for (j = 0; j < serverList[0].execArgs.size(); j++)
        serverList[i].execArgs.push_back(serverList[0].execArgs[j]);
    }

    if (serverList[i].execEnv.size() == 0) 
    {
      for (j = 0; j < serverList[0].execEnv.size(); j++)
        serverList[i].execEnv.push_back(serverList[0].execEnv[j]);
    }

    if (serverList[i].sourceFiles.size() == 0) 
    {
      for (j = 0; j < serverList[0].sourceFiles.size(); j++)
        serverList[i].sourceFiles.push_back(serverList[0].sourceFiles[j]);
    }

    if (serverList[i].inputFiles.size() == 0) 
    {
      for (j = 0; j < serverList[0].inputFiles.size(); j++)
        serverList[i].inputFiles.push_back(serverList[0].inputFiles[j]);
    }

    if (serverList[i].outputFiles.size() == 0) 
    {
      for (j = 0; j < serverList[0].outputFiles.size(); j++)
        serverList[i].outputFiles.push_back(serverList[0].outputFiles[j]);
    }

  }

  serverList.erase(serverList.begin());

  return 1;
}

///////////////////////////////////////////////////////////////////////////////
char *ProjectInfo::getErrorText()
{
  switch (errCode) 
  {
    case ERR_WRONG_VALUE      : return "Incorrect numeric value\n";
    case ERR_WRONG_PARAM      : return "Incorrect parameter\n";
    case ERR_WRONG_RUNTYPE    : return "Incorrect run type\n";
    case ERR_UNKNOWN_IDENT    : return "Unknown identifier\n";
    case ERR_UNKNOWN_VAR_NAME : return "Unknown variable\n";
    case ERR_BAD_PORT         : return "Bad port format in server's definition\n";
    case ERR_INVALID_NP       : return "Invalid np (number of processes) in server's description\n";

    default: return "No error occured\n";
  }
};

///////////////////////////////////////////////////////////////////////////////
int ProjectInfo::getVarCode(string varName)
{
  int i;

  for (i =0; i<KEYWORDS_NUM; i++)
    if (isStrEqual(varName, SCRIPT_KEYWORDS[i].name)) return SCRIPT_KEYWORDS[i].id;

  return PAR_ERROR;
}

///////////////////////////////////////////////////////////////////////////////
// Check info in the project file
int ProjectInfo::checkParams()
{
  int i;

  for (i = 0; i< serverList.size(); i++)
  {
    if (serverList[i].nProcesses == -1)
    {
      errCode = ERR_INVALID_NP;
      return 0;
    };
  }
  
	return 1;
}


